import datetime
import numpy as np
import pandas as pd

from .data import get_storage_baseline_value, \
    get_cum_capped_rb_power, get_cum_capped_qa_power

EXBI = 2**60
PIB = 2**50
LAMBDA = np.log(2) / (
    6.0 * 365
)  # minting exponential reward decay rate (6yrs half-life)
FIL_BASE = 2_000_000_000.0
STORAGE_MINING = 0.55 * FIL_BASE
SIMPLE_ALLOC = 0.3 * STORAGE_MINING  # total simple minting allocation
BASELINE_ALLOC = 0.7 * STORAGE_MINING  # total baseline minting allocation
GROWTH_RATE = float(
    np.log(2) / 365.0
)  # daily baseline growth rate (the "g" from https://spec.filecoin.io/#section-systems.filecoin_token)
BASELINE_STORAGE = (
    2.88888888
    * EXBI
    # the b_0 from https://spec.filecoin.io/#section-systems.filecoin_token
)
# TODO: understand why the baseline value from startboard differs from the spec!
# 3189227188947034973 from https://observable-api.starboard.ventures/api/v1/observable/network-storage-capacity/new_baseline_power


def compute_minting_trajectory_df(
    start_date: datetime.date,
    end_date: datetime.date,
    rb_total_power_eib: np.array,
    qa_total_power_eib: np.array,
    qa_day_onboarded_power_pib: np.array,
    qa_day_renewed_power_pib: np.array,
    minting_base: str = 'RBP'
) -> pd.DataFrame:
    # we assume minting started at main net launch, in 2020-10-15
    start_day = (start_date - datetime.date(2020, 10, 15)).days
    end_day = (end_date - datetime.date(2020, 10, 15)).days

    minting_base = minting_base.lower()
    capped_power_reference = 'network_RBP' if minting_base == 'rbp' else 'network_QAP'

    # Init dataframe
    df = pd.DataFrame(
        {
            "days": np.arange(start_day, end_day),
            "date": pd.date_range(start_date, end_date, freq="d")[:-1],
            "network_RBP": rb_total_power_eib * EXBI,
            "network_QAP": qa_total_power_eib * EXBI,
            "day_onboarded_power_QAP": qa_day_onboarded_power_pib * PIB,
            "day_renewed_power_QAP": qa_day_renewed_power_pib * PIB,
        }
    )
    df["date"] = df["date"].dt.date
    # Compute cumulative rewards due to simple minting
    df["cum_simple_reward"] = df["days"].pipe(cum_simple_minting)
    # Compute cumulative rewards due to baseline minting
    df["network_baseline"] = compute_baseline_power_array(start_date, end_date)

    df["capped_power"] = np.min(df[["network_baseline", capped_power_reference]].values, axis=1)
    zero_cum_capped_power = get_cum_capped_rb_power(start_date)
    df["cum_capped_power"] = df["capped_power"].cumsum() + zero_cum_capped_power
    df["network_time"] = df["cum_capped_power"].pipe(network_time)
    df["cum_baseline_reward"] = df["network_time"].pipe(cum_baseline_reward)
    # Add cumulative rewards and get daily rewards minted
    df["cum_network_reward"] = df["cum_baseline_reward"] + df["cum_simple_reward"]
    df["day_network_reward"] = df["cum_network_reward"].diff().fillna(method="backfill")

    return df


def cum_simple_minting(day: int) -> float:
    """
    Simple minting - the total number of tokens that should have been emitted
    by simple minting up until date provided.
    """
    return SIMPLE_ALLOC * (1 - np.exp(-LAMBDA * day))


def compute_baseline_power_array(
    start_date: datetime.date, end_date: datetime.date
) -> np.array:
    arr_len = (end_date - start_date).days
    exponents = np.arange(0, arr_len)
    init_baseline = get_storage_baseline_value(start_date)
    baseline_power_arr = init_baseline * np.exp(GROWTH_RATE * exponents)
    return baseline_power_arr


def network_time(cum_capped_power: float) -> float:
    EXA_TO_EIB = (10**18) / (2**60)
    b0 = BASELINE_STORAGE * EXA_TO_EIB
    g = GROWTH_RATE
    return (1 / g) * np.log(((g * cum_capped_power) / b0) + 1)


def cum_baseline_reward(network_time: float) -> float:
    return BASELINE_ALLOC * (1 - np.exp(-LAMBDA * network_time))



import requests
import pandas as pd
import numpy as np
import datetime
from typing import Tuple, List
import os
import json

EXBI = 2**60
PIB = 2**50

NETWORK_START = datetime.datetime(2020, 10, 15)
DEFAULT_SPACESCOPE_CHUNK_SIZE_IN_DAYS = 90
DEFAULT_AUTH_CONFIG = os.path.join(os.path.dirname(__file__), 'cfg', 'spacescope_auth.json')

class SpacescopeDataConnection:
    auth_token = ""

    def __init__(self, auth_config=None):
        if auth_config is None:
            with open(DEFAULT_AUTH_CONFIG, 'r') as f:
                default_config = json.load(f)
                SpacescopeDataConnection.auth_token = default_config['auth_key']

    @classmethod
    def spacescope_query_to_df(cls, url):
        payload={}
        headers = {
        'authorization': cls.auth_token
        }
        response = requests.request("GET", url, headers=headers, data=payload)
        df = pd.DataFrame(response.json()['data'])
        return df

    @staticmethod
    def get_historical_network_stats(
        start_date: datetime.date, current_date: datetime.date, end_date: datetime.date
    ) -> pd.DataFrame:
        power_df = SpacescopeDataConnection.query_spacescope_power_stats(start_date, current_date)
        onboards_df = SpacescopeDataConnection.query_spacescope_daily_power_onboarded(start_date, current_date)
        stats_df = SpacescopeDataConnection.query_spacescope_supply_stats(start_date, current_date)
        stats_df = stats_df.merge(power_df, on="date", how="inner").merge(
            onboards_df, on="date", how="inner"
        )
        renewal_df = SpacescopeDataConnection.get_day_renewed_power_stats(
            start_date, current_date, end_date
        )
        stats_df = stats_df.merge(renewal_df, on='date', how='inner')
        return stats_df

    @staticmethod
    def get_sector_expiration_stats(
        start_date: datetime.date,
        current_date: datetime.date,
        end_date: datetime.date,
    ) -> pd.DataFrame:
        scheduled_df = SpacescopeDataConnection.query_spacescope_sector_expirations(start_date, end_date)
        filter_scheduled_df = scheduled_df[
            scheduled_df["date"] >= pd.to_datetime(current_date, utc="UTC")
        ]
        rbp_expire_vec = filter_scheduled_df["total_rb"].values
        qap_expire_vec = filter_scheduled_df["total_qa"].values
        # we need the entire history of known_scheduled_pledge_release, so get the
        # data from the entire time-window, not just from current-date onwards
        pledge_release_vec = scheduled_df["total_pledge"].values
        return rbp_expire_vec, qap_expire_vec, pledge_release_vec

    @staticmethod
    def get_day_renewed_power_stats(
        start_date: datetime.date,
        current_date: datetime.date,
        end_date: datetime.date,
    ) -> Tuple[np.array, np.array]:
        scheduled_df = SpacescopeDataConnection.query_spacescope_sector_expirations(start_date, end_date)
        filter_scheduled_df = scheduled_df[
            scheduled_df["date"] <= pd.to_datetime(current_date, utc="UTC")
        ]
        rb_renewal_rate = (
            filter_scheduled_df["extended_rb"] / filter_scheduled_df["total_rb"]
        ).values
        day_renewed_qa_power = filter_scheduled_df["extended_qa"].values
        renewal_df = pd.DataFrame(
            {
                'date': pd.to_datetime(filter_scheduled_df['date']).dt.date,
                'rb_renewal_rate': rb_renewal_rate,
                'day_renewed_qa_power_pib': day_renewed_qa_power
            }
        )
        return renewal_df

    @staticmethod
    def chunk_dates(start_date: datetime.date, 
                    end_date: datetime.date,
                    chunks_days: int = DEFAULT_SPACESCOPE_CHUNK_SIZE_IN_DAYS) -> List:
        chunk_start = start_date
        dates_chunked = []
        while chunk_start <= end_date:
            chunk_end = min(chunk_start + datetime.timedelta(days=chunks_days), end_date)
            dates_chunked.append((chunk_start, chunk_end))

            chunk_start = chunk_end + datetime.timedelta(days=1)

        return dates_chunked

    @staticmethod
    def query_spacescope_sector_expirations(
        start_date: datetime.date, 
        end_date: datetime.date,
    ) -> pd.DataFrame:
        # See: https://docs.spacescope.io/network_core/power/#request-url-4
        #  NOTE: this is a bit weird compared to the rest of the Spacescope API, where scheduled expirations
        #  does not need a start/end date and returns the entire dataset.  For now, we use this and filter
        #  but this may need to change in the future if Spacescope changes their API.
        url = "https://api.spacescope.io/v2/power/sectors_schedule_expiration"
        scheduled_df = SpacescopeDataConnection.spacescope_query_to_df(url)
        
        # Convert bytes to pebibytes
        scheduled_df["extended_rb"] = scheduled_df["extended_bytes"].astype(float) / PIB
        scheduled_df["expired_rb"] = scheduled_df["expired_bytes"].astype(float) / PIB
        scheduled_df["terminated_rb"] = scheduled_df["terminated_bytes"].astype(float) / PIB
        scheduled_df['schedule_expire_rb'] = scheduled_df["schedule_expire_bytes"].astype(float) / PIB

        scheduled_df["extended_qa"] = scheduled_df["extended_bytes_qap"].astype(float) / PIB
        scheduled_df["expired_qa"] = scheduled_df["expired_bytes_qap"].astype(float) / PIB
        scheduled_df['terminated_qa'] = scheduled_df['terminated_bytes_qap'].astype(float) / PIB
        scheduled_df["schedule_expire_qa"] = scheduled_df["schedule_expire_bytes_qap"].astype(float) / PIB

        scheduled_df["extended_pledge"] = scheduled_df["extended_pledge"].astype(float)
        scheduled_df["expired_pledge"] = scheduled_df["expired_pledge"].astype(float)
        scheduled_df["terminated_pledge"] = scheduled_df["terminated_pledge"].astype(float)
        scheduled_df["schedule_expire_pledge"] = scheduled_df["schedule_expire_pledge"].astype(float)

        # Total scheduled to expire, excluding terminated. Exclude terminated because 
        scheduled_df["total_rb"] = (
            scheduled_df["schedule_expire_rb"] - scheduled_df['terminated_rb']
        )
        scheduled_df["total_qa"] = (
            scheduled_df["schedule_expire_qa"] - scheduled_df['terminated_qa']
        )
        scheduled_df["total_pledge"] = (
            scheduled_df["schedule_expire_pledge"] - scheduled_df['terminated_pledge']
        )
        # Convert interest date to datetime
        scheduled_df["date"] = pd.to_datetime(scheduled_df["interest_date"])
        # Filter dates
        scheduled_df = scheduled_df[
            scheduled_df["date"] >= pd.to_datetime(start_date, utc="UTC")
        ]
        scheduled_df = scheduled_df[
            scheduled_df["date"] < pd.to_datetime(end_date, utc="UTC")
        ]
        return scheduled_df

    @staticmethod
    def query_spacescope_daily_power_onboarded(
        start_date: datetime.date, 
        end_date: datetime.date,
        chunk_days: int = DEFAULT_SPACESCOPE_CHUNK_SIZE_IN_DAYS
    ) -> pd.DataFrame:
        url_template = "https://api.spacescope.io/v2/power/daily_power_onboarding_by_sector_size?end_date=%s&start_date=%s"
        df = SpacescopeDataConnection.spacescope_query(start_date, end_date, url_template, chunk_days)
        df['day_onboarded_rb_power_pib'] = (df['commit_rbp_32gib'] + df['commit_rbp_64gib']) / PIB
        df['day_onboarded_qa_power_pib'] = (df['commit_qap_32gib'] + df['commit_qap_64gib']) / PIB
        df['date'] = pd.to_datetime(df['stat_date']).dt.date
        
        # Filter columns
        onboards_df = df[
            ["date", "day_onboarded_rb_power_pib", "day_onboarded_qa_power_pib"]
        ]
        return onboards_df

    @staticmethod
    def query_spacescope_supply_stats(
        start_date: datetime.date,
        end_date: datetime.date,
        chunk_days: int = DEFAULT_SPACESCOPE_CHUNK_SIZE_IN_DAYS
    ) -> pd.DataFrame:
        url_template = "https://api.spacescope.io/v2/circulating_supply/circulating_supply?end_date=%s&start_date=%s"
        raw_stats_df = SpacescopeDataConnection.spacescope_query(start_date, end_date, url_template, chunk_days)
        # Convert metrics to float
        stats_df = raw_stats_df[
            [
                "circulating_fil",
                "mined_fil",
                "vested_fil",
                "locked_fil",
                "burnt_fil",
                "reserve_disbursed_fil"
            ]
        ].astype(float)
        # Convert dates to datetime dates
        stats_df["date"] = pd.to_datetime(raw_stats_df["stat_date"]).dt.date
        # Filter dates
        stats_df = stats_df[
            (stats_df["date"] >= start_date) & (stats_df["date"] <= end_date)
        ]
        return stats_df

    @staticmethod
    def spacescope_query(start_date: datetime.date, 
                        end_date: datetime.date,
                        url_template: str,
                        chunk_days: int = DEFAULT_SPACESCOPE_CHUNK_SIZE_IN_DAYS) -> pd.DataFrame:
        dates_chunked = SpacescopeDataConnection.chunk_dates(start_date, end_date, chunks_days=chunk_days)
        df_list = []
        for d in dates_chunked:
            chunk_start = d[0].strftime('%Y-%m-%d')
            chunk_end = d[1].strftime('%Y-%m-%d')
            url = url_template % (chunk_end, chunk_start)
            df = SpacescopeDataConnection.spacescope_query_to_df(url)
            df_list.append(df)
        
        df_all = pd.concat(df_list, ignore_index=True)
        return df_all

    @staticmethod
    def query_historical_power(start_date: datetime.date, 
                            end_date: datetime.date,
                            chunk_days: int = DEFAULT_SPACESCOPE_CHUNK_SIZE_IN_DAYS) -> pd.DataFrame:
        url_template = "https://api.spacescope.io/v2/power/network_storage_capacity?end_date=%s&start_date=%s"
        df = SpacescopeDataConnection.spacescope_query(start_date, end_date, url_template, chunk_days)
        
        df['date'] = pd.to_datetime(df['stat_date']).dt.date
        df['total_qa_bytes_power'] = df['total_qa_bytes_power'].astype(float)
        df['total_raw_bytes_power'] = df['total_raw_bytes_power'].astype(float)
        df['baseline_power'] = df['baseline_power'].astype(float)

        return df

    @staticmethod
    def query_spacescope_power_stats(
        start_date: datetime.date, end_date: datetime.date,
        chunk_days: int = DEFAULT_SPACESCOPE_CHUNK_SIZE_IN_DAYS
    ) -> pd.DataFrame:
        power_df = SpacescopeDataConnection.query_historical_power(
            start_date, end_date, chunk_days=chunk_days
        )
        # Convert power stats to exibytes
        power_df["total_raw_power_eib"] = (
            power_df["total_raw_bytes_power"] / EXBI
        )
        power_df["total_qa_power_eib"] = (
            power_df["total_qa_bytes_power"] / EXBI
        )
        # Select final columns
        power_df = power_df[["date", "total_raw_power_eib", "total_qa_power_eib"]]
        return power_df

    @staticmethod
    def query_historical_baseline_power(start_date: datetime.date = None,
                                end_date: datetime.date = None,
                                chunk_days: int = DEFAULT_SPACESCOPE_CHUNK_SIZE_IN_DAYS) -> pd.DataFrame:
        if start_date is None:
            start_date = NETWORK_START
        if end_date is None:
            end_date = datetime.datetime.today()

        historical_power_df = SpacescopeDataConnection.query_historical_power(
            start_date, end_date, chunk_days=chunk_days
        )

        bp_df = historical_power_df[['date', 'baseline_power']]
        bp_df = bp_df.rename(columns={'baseline_power': 'baseline'})
        return bp_df
        
    @staticmethod
    def query_historical_rb_power(start_date: datetime.date = None,
                                end_date: datetime.date = None,
                                chunk_days: int = DEFAULT_SPACESCOPE_CHUNK_SIZE_IN_DAYS) -> pd.DataFrame:
        if start_date is None:
            start_date = NETWORK_START
        if end_date is None:
            end_date = datetime.datetime.today()

        historical_power_df = SpacescopeDataConnection.query_historical_power(
            start_date, end_date, chunk_days=chunk_days
        )

        rbp_df = historical_power_df[['date', 'total_raw_bytes_power']]
        rbp_df = rbp_df.rename(columns={'total_raw_bytes_power': 'rb_power'})
        return rbp_df

    @staticmethod
    def query_historical_qa_power(start_date: datetime.date = None,
                                end_date: datetime.date = None,
                                chunk_days: int = DEFAULT_SPACESCOPE_CHUNK_SIZE_IN_DAYS) -> pd.DataFrame:
        if start_date is None:
            start_date = NETWORK_START
        if end_date is None:
            end_date = datetime.datetime.today()

        historical_power_df = SpacescopeDataConnection.query_historical_power(
            start_date, end_date, chunk_days=chunk_days
        )
        qap_df = historical_power_df[['date', 'total_raw_bytes_power']]
        qap_df = qap_df.rename(columns={'total_qa_bytes_power': 'qa_power'})

        return qap_df



import numpy as np
import pandas as pd
import datetime
from typing import Union

from .locking import (
    get_day_schedule_pledge_release,
    compute_day_reward_release,
    compute_day_delta_pledge,
    compute_day_locked_rewards,
    compute_day_locked_pledge,
)
from .power import scalar_or_vector_to_vector

"""
There is still a small discrepancy between the actual locked FIL and forecasted
locked FIL. We believe that it could be due to the following reasons:
  a) Sector durations are not unlocked after exactly 1y. In general they're distributed and slightly longer. But in that case I’d expect the sign of the error to be the opposite to observed.
  b) The error between actual and forecasted locked FIL is 0 for day_idx=1. This might imply that a build up of errors due to an error in `day_locked_pledge` sounds more like it could be the issue.
  c) If we're sure the locking discrepancy is not a bug but rather a deficiency in the model popping up via the approximations used, we may way want to include a learnable factor to correct the difference
"""


def forecast_circulating_supply_df(
    start_date: datetime.date,
    current_date: datetime.date,
    end_date: datetime.date,
    circ_supply_zero: float,
    locked_fil_zero: float,
    daily_burnt_fil: float,
    duration: int,
    renewal_rate: Union[np.array, float],
    burnt_fil_vec: np.array,
    vest_df: pd.DataFrame,
    mint_df: pd.DataFrame,
    known_scheduled_pledge_release_vec: np.array,
    lock_target: float = 0.3,
) -> pd.DataFrame:
    # we assume all stats started at main net launch, in 2020-10-15
    start_day = (start_date - datetime.date(2020, 10, 15)).days
    current_day = (current_date - datetime.date(2020, 10, 15)).days
    end_day = (end_date - datetime.date(2020, 10, 15)).days
    # initialise dataframe and auxilialy variables
    df = initialise_circulating_supply_df(
        start_date,
        end_date,
        circ_supply_zero,
        locked_fil_zero,
        burnt_fil_vec,
        vest_df,
        mint_df,
    )
    circ_supply = circ_supply_zero
    sim_len = end_day - start_day
    renewal_rate_vec = scalar_or_vector_to_vector(renewal_rate, sim_len)

    # Simulation for loop
    current_day_idx = current_day - start_day
    for day_idx in range(1, sim_len):
        # Compute daily change in initial pledge collateral
        day_pledge_locked_vec = df["day_locked_pledge"].values
        scheduled_pledge_release = get_day_schedule_pledge_release(
            day_idx,
            current_day_idx,
            day_pledge_locked_vec,
            known_scheduled_pledge_release_vec,
            duration,
        )
        pledge_delta = compute_day_delta_pledge(
            df["day_network_reward"].iloc[day_idx],
            circ_supply,
            df["day_onboarded_power_QAP"].iloc[day_idx],
            df["day_renewed_power_QAP"].iloc[day_idx],
            df["network_QAP"].iloc[day_idx],
            df["network_baseline"].iloc[day_idx],
            renewal_rate_vec[day_idx],
            scheduled_pledge_release,
            lock_target,
        )
        # Get total locked pledge (needed for future day_locked_pledge)
        day_locked_pledge, day_renewed_pledge = compute_day_locked_pledge(
            df["day_network_reward"].iloc[day_idx],
            circ_supply,
            df["day_onboarded_power_QAP"].iloc[day_idx],
            df["day_renewed_power_QAP"].iloc[day_idx],
            df["network_QAP"].iloc[day_idx],
            df["network_baseline"].iloc[day_idx],
            renewal_rate_vec[day_idx],
            scheduled_pledge_release,
            lock_target,
        )
        # Compute daily change in block rewards collateral
        day_locked_rewards = compute_day_locked_rewards(
            df["day_network_reward"].iloc[day_idx]
        )
        day_reward_release = compute_day_reward_release(
            df["network_locked_reward"].iloc[day_idx - 1]
        )
        reward_delta = day_locked_rewards - day_reward_release
        # Update dataframe
        df["day_locked_pledge"].iloc[day_idx] = day_locked_pledge
        df["day_renewed_pledge"].iloc[day_idx] = day_renewed_pledge
        df["network_locked_pledge"].iloc[day_idx] = (
            df["network_locked_pledge"].iloc[day_idx - 1] + pledge_delta
        )
        df["network_locked_reward"].iloc[day_idx] = (
            df["network_locked_reward"].iloc[day_idx - 1] + reward_delta
        )
        df["network_locked"].iloc[day_idx] = (
            df["network_locked"].iloc[day_idx - 1] + pledge_delta + reward_delta
        )
        # Update gas burnt
        if df["network_gas_burn"].iloc[day_idx] == 0.0:
            df["network_gas_burn"].iloc[day_idx] = (
                df["network_gas_burn"].iloc[day_idx - 1] + daily_burnt_fil
            )
        # Find circulating supply balance and update
        circ_supply = (
            df["disbursed_reserve"].iloc[
                day_idx
            ]  # from initialise_circulating_supply_df
            + df["cum_network_reward"].iloc[day_idx]  # from the minting_model
            + df["total_vest"].iloc[day_idx]  # from vesting_model
            - df["network_locked"].iloc[day_idx]  # from simulation loop
            - df["network_gas_burn"].iloc[day_idx]  # comes from user inputs
        )
        df["circ_supply"].iloc[day_idx] = max(circ_supply, 0)
    return df


def initialise_circulating_supply_df(
    start_date: datetime.date,
    end_date: datetime.date,
    circ_supply_zero: float,
    locked_fil_zero: float,
    burnt_fil_vec: np.array,
    vest_df: pd.DataFrame,
    mint_df: pd.DataFrame,
) -> pd.DataFrame:
    # we assume days start at main net launch, in 2020-10-15
    start_day = (start_date - datetime.date(2020, 10, 15)).days
    end_day = (end_date - datetime.date(2020, 10, 15)).days
    len_sim = end_day - start_day
    df = pd.DataFrame(
        {
            "days": np.arange(start_day, end_day),
            "date": pd.date_range(start_date, end_date, freq="d")[:-1],
            "circ_supply": np.zeros(len_sim),
            "network_gas_burn": np.pad(
                burnt_fil_vec, (0, len_sim - len(burnt_fil_vec))
            ),
            "day_locked_pledge": np.zeros(len_sim),
            "day_renewed_pledge": np.zeros(len_sim),
            "network_locked_pledge": np.zeros(len_sim),
            "network_locked": np.zeros(len_sim),
            "network_locked_reward": np.zeros(len_sim),
            "disbursed_reserve": np.ones(len_sim)
            * (17066618961773411890063046 * 10**-18),
        }
    )
    df["date"] = df["date"].dt.date
    df["network_locked_pledge"].iloc[0] = locked_fil_zero / 2.0
    df["network_locked_reward"].iloc[0] = locked_fil_zero / 2.0
    df["network_locked"].iloc[0] = locked_fil_zero
    df["circ_supply"].iloc[0] = circ_supply_zero
    df = df.merge(vest_df, on="date", how="inner")
    df = df.merge(mint_df.drop(columns=["days"]), on="date", how="inner")
    return df



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mechaFIL",
    version="1.4",
    author="Maria Silva, Tom Mellan, Kiran Karra",
    author_email="misilva73@gmail.com, t.mellan@imperial.ac.uk, kiran.karra@gmail.com",
    description="Mechanistic model for the Filecoin Economy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    project_urls={
        "Documentation": "https://github.com/protocol/filecoin-mecha-twin",
        "Source": "https://github.com/protocol/filecoin-mecha-twin",
    },
    packages=['mechafil'],
    package_data={'mechafil': ['cfg/spacescope_auth.json']},
    install_requires=["numpy>=1.23.1", "pandas>=1.4.3", "requests>=2.28.1"],
    python_requires=">=3.8",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import numpy as np
import pandas as pd
import datetime
from typing import Callable, Tuple, Union
import numbers

from .utils import validate_qap_method

# --------------------------------------------------------------------------------------
#  Utility functions
# --------------------------------------------------------------------------------------
def scalar_or_vector_to_vector(
    input_x: Union[np.array, float], expected_len: int, err_msg: str = None
) -> np.array:
    if isinstance(input_x, numbers.Number):
        return np.ones(expected_len) * input_x
    else:
        err_msg_out = (
            "vector input does not match expected length!"
            if err_msg is None
            else err_msg
        )
        assert len(input_x) == expected_len, err_msg_out
        return input_x


# --------------------------------------------------------------------------------------
#  QA Multiplier functions
# --------------------------------------------------------------------------------------
def compute_qa_factor(
    fil_plus_rate: Union[np.array, float],
    fil_plus_m: float = 10.0,
    duration_m: Callable = None,
    duration: int = None,
) -> Union[np.array, float]:
    fil_plus_multipler = 1.0 + (fil_plus_m - 1) * fil_plus_rate
    if duration_m is None:
        return fil_plus_multipler
    else:
        return duration_m(duration) * fil_plus_multipler


# --------------------------------------------------------------------------------------
#  Onboardings
# --------------------------------------------------------------------------------------
def forecast_rb_daily_onboardings(
    rb_onboard_power: float, forecast_lenght: int
) -> np.array:
    rb_onboarded_power_vec = scalar_or_vector_to_vector(
        rb_onboard_power,
        forecast_lenght,
        err_msg="If rb_onboard_power is provided as a vector, it must be the same length as the forecast length",
    )
    return rb_onboarded_power_vec


def forecast_qa_daily_onboardings(
    rb_onboard_power: Union[np.array, float],
    fil_plus_rate: Union[np.array, float],
    forecast_lenght: int,
    fil_plus_m: float = 10.0,
    duration_m: Callable = None,
    duration: int = None,
) -> np.array:
    # If duration_m is not provided, qa_factor = 1.0 + 9.0 * fil_plus_rate
    qa_factor = compute_qa_factor(fil_plus_rate, fil_plus_m, duration_m, duration)
    qa_onboard_power = qa_factor * rb_onboard_power
    qa_onboard_power_vec = scalar_or_vector_to_vector(
        qa_onboard_power,
        forecast_lenght,
        err_msg="If qa_onboard_power is provided as a vector, it must be the same length as the forecast length",
    )
    return qa_onboard_power_vec


# --------------------------------------------------------------------------------------
#  Renewals
# --------------------------------------------------------------------------------------
def compute_basic_day_renewed_power(
    day_i: int,
    day_scheduled_expire_power_vec: np.array,
    renewal_rate_vec: np.array,
) -> float:
    day_renewed_power = renewal_rate_vec[day_i] * day_scheduled_expire_power_vec[day_i]
    return day_renewed_power


def compute_day_qa_renewed_power(
    day_i: int,
    day_rb_scheduled_expire_power_vec: np.array,
    renewal_rate_vec: np.array,
    fil_plus_rate: float,
    fil_plus_m: float = 10.0,
    duration_m: Callable = None,
    duration: int = None,
) -> float:
    fpr = (
        fil_plus_rate
        if isinstance(fil_plus_rate, numbers.Number)
        else fil_plus_rate[day_i]
    )
    qa_factor = compute_qa_factor(fpr, fil_plus_m, duration_m, duration)
    day_renewed_power = (
        qa_factor * renewal_rate_vec[day_i] * day_rb_scheduled_expire_power_vec[day_i]
    )
    return day_renewed_power


# --------------------------------------------------------------------------------------
#  Scheduled expirations
# --------------------------------------------------------------------------------------
def compute_day_se_power(
    day_i: int,
    known_scheduled_expire_vec: np.array,
    day_onboard_vec: np.array,
    day_renewed_vec: np.array,
    duration: int,
) -> float:
    # Scheduled expirations coming from known active sectors
    if day_i > len(known_scheduled_expire_vec) - 1:
        known_day_se_power = 0.0
    else:
        known_day_se_power = known_scheduled_expire_vec[day_i]
    # Scheduled expirations coming from modeled sectors
    if day_i - duration >= 0:
        model_day_se_power = (
            day_onboard_vec[day_i - duration] + day_renewed_vec[day_i - duration]
        )
    else:
        model_day_se_power = 0.0
    # Total scheduled expirations
    day_se_power = known_day_se_power + model_day_se_power
    return day_se_power


# --------------------------------------------------------------------------------------
#  Forecast Power stats
# --------------------------------------------------------------------------------------
def forecast_power_stats(
    rb_power_zero: float,
    qa_power_zero: float,
    rb_onboard_power: Union[np.array, float],
    rb_known_scheduled_expire_vec: np.array,
    qa_known_scheduled_expire_vec: np.array,
    renewal_rate: Union[np.array, float],
    fil_plus_rate: Union[np.array, float],
    duration: int,
    forecast_lenght: int,
    fil_plus_m: float = 10.0,
    duration_m: Callable = None,
    qap_method: str = 'tunable'  # can be set to tunable or basic
                                 # see: https://hackmd.io/O6HmAb--SgmxkjLWSpbN_A?view
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    validate_qap_method(qap_method)
    
    # Forecast onboards
    renewal_rate_vec = scalar_or_vector_to_vector(
        renewal_rate,
        forecast_lenght,
        err_msg="If renewal_rate is provided as a vector, it must be the same length as the forecast length",
    )

    day_rb_onboarded_power = forecast_rb_daily_onboardings(
        rb_onboard_power, forecast_lenght
    )

    total_rb_onboarded_power = day_rb_onboarded_power.cumsum()
    day_qa_onboarded_power = forecast_qa_daily_onboardings(
        rb_onboard_power,
        fil_plus_rate,
        forecast_lenght,
        fil_plus_m,
        duration_m,
        duration,
    )
    total_qa_onboarded_power = day_qa_onboarded_power.cumsum()
    # Initialize scheduled expirations and renewals
    day_rb_scheduled_expire_power = np.zeros(forecast_lenght)
    day_rb_renewed_power = np.zeros(forecast_lenght)
    day_qa_scheduled_expire_power = np.zeros(forecast_lenght)
    day_qa_renewed_power = np.zeros(forecast_lenght)
    # Run loop to forecast daily scheduled expirations and renewals
    for day_i in range(forecast_lenght):
        # Raw-power stats
        day_rb_scheduled_expire_power[day_i] = compute_day_se_power(
            day_i,
            rb_known_scheduled_expire_vec,
            day_rb_onboarded_power,
            day_rb_renewed_power,
            duration,
        )
        day_rb_renewed_power[day_i] = compute_basic_day_renewed_power(
            day_i, day_rb_scheduled_expire_power, renewal_rate_vec
        )
        # Quality-adjusted stats
        day_qa_scheduled_expire_power[day_i] = compute_day_se_power(
            day_i,
            qa_known_scheduled_expire_vec,
            day_qa_onboarded_power,
            day_qa_renewed_power,
            duration,
        )
        
        # see https://hackmd.io/O6HmAb--SgmxkjLWSpbN_A?view for more details
        if qap_method == 'tunable':
            day_qa_renewed_power[day_i] = compute_day_qa_renewed_power(
                day_i,
                day_rb_scheduled_expire_power,
                renewal_rate_vec,
                fil_plus_rate,
                fil_plus_m,
                duration_m,
                duration,
            )
        elif qap_method == 'basic':
            day_qa_renewed_power[day_i] = compute_basic_day_renewed_power(
                day_i, day_qa_scheduled_expire_power, renewal_rate_vec
            )
    # Compute total scheduled expirations and renewals
    total_rb_scheduled_expire_power = day_rb_scheduled_expire_power.cumsum()
    total_rb_renewed_power = day_rb_renewed_power.cumsum()
    total_qa_scheduled_expire_power = day_qa_scheduled_expire_power.cumsum()
    total_qa_renewed_power = day_qa_renewed_power.cumsum()
    # Total RB power
    rb_power_zero_vec = np.ones(forecast_lenght) * rb_power_zero
    rb_total_power = (
        rb_power_zero_vec
        + total_rb_onboarded_power
        - total_rb_scheduled_expire_power
        + total_rb_renewed_power
    )
    # Total QA power
    qa_power_zero_vec = np.ones(forecast_lenght) * qa_power_zero
    qa_total_power = (
        qa_power_zero_vec
        + total_qa_onboarded_power
        - total_qa_scheduled_expire_power
        + total_qa_renewed_power
    )
    # Build DataFrames
    rb_df = pd.DataFrame(
        {
            "forecasting_step": np.arange(forecast_lenght),
            "onboarded_power": day_rb_onboarded_power,
            "cum_onboarded_power": total_rb_onboarded_power,
            "expire_scheduled_power": day_rb_scheduled_expire_power,
            "cum_expire_scheduled_power": total_rb_scheduled_expire_power,
            "renewed_power": day_rb_renewed_power,
            "cum_renewed_power": total_rb_renewed_power,
            "total_power": rb_total_power,
        }
    )
    rb_df["power_type"] = "raw-byte"
    qa_df = pd.DataFrame(
        {
            "forecasting_step": np.arange(forecast_lenght),
            "onboarded_power": day_qa_onboarded_power,
            "cum_onboarded_power": total_qa_onboarded_power,
            "expire_scheduled_power": day_qa_scheduled_expire_power,
            "cum_expire_scheduled_power": total_qa_scheduled_expire_power,
            "renewed_power": day_qa_renewed_power,
            "cum_renewed_power": total_qa_renewed_power,
            "total_power": qa_total_power,
        }
    )
    qa_df["power_type"] = "quality-adjusted"
    return rb_df, qa_df


# --------------------------------------------------------------------------------------
#  Build power stats DataFrame
# --------------------------------------------------------------------------------------
def build_full_power_stats_df(
    stats_df: pd.DataFrame,
    rb_power_df: pd.DataFrame,
    qa_power_df: pd.DataFrame,
    start_date: datetime.date,
    current_date: datetime.date,
    end_date: datetime.date,
) -> pd.DataFrame:
    # Past power
    past_power_df = stats_df[
        [
            "date",
            "total_raw_power_eib",
            "total_qa_power_eib",
            "day_onboarded_qa_power_pib",
            "day_renewed_qa_power_pib",
        ]
    ]
    # Forecasted power
    forecast_power_df = rb_power_df[["total_raw_power_eib"]]
    forecast_power_df.loc[:, "total_qa_power_eib"] = qa_power_df["total_qa_power_eib"]
    forecast_power_df.loc[:, "day_onboarded_qa_power_pib"] = qa_power_df[
        "onboarded_power"
    ]
    forecast_power_df.loc[:, "day_renewed_qa_power_pib"] = qa_power_df["renewed_power"]
    forecast_start_date = current_date + datetime.timedelta(days=1)
    forecast_power_df.loc[:, "date"] = pd.date_range(
        start=forecast_start_date, end=end_date, freq="d"
    )
    forecast_power_df["date"] = forecast_power_df["date"].dt.date
    # All power stats
    concat_df = pd.concat([past_power_df, forecast_power_df])
    power_df = pd.DataFrame(
        {"date": pd.date_range(start=start_date, end=end_date, freq="d")}
    )
    power_df["date"] = power_df["date"].dt.date
    power_df = power_df.merge(concat_df, on="date", how="left").fillna(
        method="backfill"
    )
    power_df = power_df.iloc[:-1]
    return power_df
